import sys
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from .checks import has_annotated_init
from .checks import is_list
from .checks import is_mapping
from .checks import is_union
from .compat import ForwardRef
from .dispatchers.multi import MultiDispatcher
from .forges import _clean_annotations
from .forges import forge_annotated_init
from .forges import forge_list
from .forges import forge_mapping
from .forges import forge_none
from .forges import forge_union
from .types import CannotDispatch
from .types import Check
from .types import Forge
from .types import NoneType

T = TypeVar("T")


class Svarog:
    _refs_owners: Dict[ForwardRef, Type]

    def __init__(self):
        self._refs_owners = {}
        self._dispatcher = MultiDispatcher()

        self._dispatcher.register_cls(NoneType, forge_none)
        self._dispatcher.register_func(has_annotated_init, forge_annotated_init)

        self._dispatcher.register_func(is_list, forge_list)
        self._dispatcher.register_func(is_mapping, forge_mapping)
        self._dispatcher.register_func(is_union, forge_union)

    def register_forge(self, type_: Type, forge: Forge) -> None:
        self._dispatcher.register_cls(type_, forge)

    def register_mold(self, check: Check, forge: Forge) -> None:
        self._dispatcher.register_func(check, forge)

    @lru_cache(None)
    def _update_forward_ref_owners_registry(self, type_: Type) -> None:
        def update_subtype(sub_type: Type) -> None:
            if isinstance(sub_type, ForwardRef):
                self._refs_owners[sub_type] = type_
            if (
                not isinstance(sub_type, NoneType)  # type: ignore
                and sub_type is not Any
                and hasattr(sub_type, "__args__")
            ):
                for sub_type in annotation.__args__:
                    update_subtype(sub_type)

        for _, annotation in _clean_annotations(type_.__init__):
            update_subtype(annotation)

    def forge(self, type_: Type[T], data: Any) -> T:
        if isinstance(type_, ForwardRef):
            type_ = self._resolve_forward_ref(type_)

        if has_annotated_init(type_):
            self._update_forward_ref_owners_registry(type_)

        if type_ is Any:
            return data

        handler = self._dispatcher.dispatch(type_)
        try:
            return handler(type_, data, self.forge)
        except CannotDispatch:
            return type_(data)  # type: ignore

    def _resolve_forward_ref(self, type_: ForwardRef) -> Type:
        try:
            owner: Type = self._refs_owners[type_]
        except KeyError:
            raise RuntimeError(f"Unknown forward ref: {type_}")
        return type_._evaluate(  # type: ignore
            vars(sys.modules[owner.__module__]), getattr(owner, "__dict__", {})
        )



import pytest

from svarog import Svarog
from svarog.types import Forge


@pytest.fixture
def svarog() -> Svarog:
    return Svarog()


@pytest.fixture
def forge(svarog) -> Forge:
    return svarog.forge


@pytest.fixture
def register_forge(svarog):
    return svarog.register_forge



from unittest.mock import Mock

from svarog.dispatchers.functional import FunctionalDispatch
from svarog.dispatchers.multi import MultiDispatcher


def test_functional_dispatcher_should_call_default_handler():
    default = Mock()
    type = Mock()
    data = Mock()
    forge = Mock()
    fb = FunctionalDispatch(default)

    fb(type, data, forge)
    default.assert_called_once_with(type, data, forge)


def test_functional_dispatcher_should_call_handler_based_on_check():
    default = Mock()
    odd_handler = Mock()
    even_handler = Mock()
    forge = Mock()
    fb = FunctionalDispatch(default)

    def is_odd(type):
        return type % 2 == 1

    def is_even(type):
        return type % 2 == 0

    fb.register(is_odd)(odd_handler)
    fb.register(is_even)(even_handler)

    fb(2, "foo", forge)
    fb(1, "bar", forge)

    even_handler.assert_called_once_with(2, "foo", forge)
    odd_handler.assert_called_once_with(1, "bar", forge)

    default.assert_not_called()


def test_multi_dispatcher_will_dispatch_for_registered_type():
    md = MultiDispatcher()
    handle_class = Mock()
    forge = Mock()

    class A:
        pass

    class B(A):
        pass

    md.register_cls(A, handle_class)
    md.dispatch(B)(B, "data", forge)

    handle_class.assert_called_once_with(B, "data", forge)


def test_multi_dispatcher_will_dispatch_for_registered_func():
    md = MultiDispatcher()
    handle_func = Mock()
    forge = Mock()

    class A:
        pass

    def is_a(type):
        return type == A

    md.register_func(is_a, handle_func)
    md.dispatch(A)(A, "data", forge)

    handle_func.assert_called_once_with(A, "data", forge)



from typing import _GenericAlias  # type: ignore
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import MutableSequence
from typing import Sequence
from typing import Type
from typing import TypeVar
from typing import Union

from .compat import _SpecialGenericAlias
from .compat import get_args
from .compat import get_origin

T = TypeVar("T")


def has_annotated_init(type_: Type[T]) -> bool:
    return hasattr(type_, "__init__") and hasattr(type_.__init__, "__annotations__")


def is_union(type_: Union) -> bool:
    return (
        type_ is Union or isinstance(type_, _GenericAlias) and type_.__origin__ is Union
    )


def is_list(type_: Any) -> bool:
    return type_ is List or (
        isinstance(type_, (_GenericAlias, _SpecialGenericAlias))
        and issubclass(get_origin(type_), Sequence)  # type: ignore
    )


def is_bare(type_: Any) -> bool:
    args = get_args(type_)
    return (
        args == get_args(List)
        or args == get_args(Sequence)
        or args == get_args(Mapping)
        or args == get_args(Dict)
        or args == get_args(MutableSequence)
    )


def is_mapping(type_: Any) -> bool:
    return type_ is Mapping or (
        type_.__class__ is _GenericAlias and issubclass(type_.__origin__, Mapping)
    )



import inspect
from functools import lru_cache
from typing import Any
from typing import ForwardRef
from typing import List
from typing import Mapping
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from .checks import is_bare
from .compat import get_args
from .types import CannotDispatch
from .types import Forge
from .types import NoneType


T = TypeVar("T")


@lru_cache()
def _cached_defaults(thingy):
    parameters = inspect.signature(thingy).parameters
    return {
        name: parameter.default
        for name, parameter in parameters.items()
        if parameter.default != inspect._empty  # type: ignore
    }


def _clean_annotations(annotated):
    if annotated.__annotations__ == {}:
        for varname in annotated.__code__.co_varnames:
            if varname != "self":
                yield varname, Any
    else:
        for name, value in annotated.__annotations__.items():
            if isinstance(value, str):
                yield name, ForwardRef(value)
            else:
                yield name, value


def forge_none(type_: NoneType, data: Any, forge: Forge) -> None:
    return None


def forge_annotated_init(type_: Type[T], data: Any, forge: Forge) -> T:
    if not hasattr(data, "items"):
        raise CannotDispatch()

    forged = {
        name: forge(value, data[name])
        for name, value in _clean_annotations(type_.__init__)
        if name != "return" and name in data
    }
    kwargs = dict(_cached_defaults(type_.__init__))
    kwargs.update(forged)

    return type_(**kwargs)  # type: ignore


def forge_union(union: Union, data: Any, forge: Forge) -> Optional[T]:
    params = union.__args__  # type: ignore
    if NoneType in params:
        if data is None:
            return None

        if len(params) == 2:
            first, second = params
            other = first if second is NoneType else second
            return forge(other, data)

    raise NotImplementedError("Unions other as optionals are not supported yet")


def forge_list(type_: Type[List], list_: List, forge: Forge) -> List:
    if is_bare(type_):
        return list(list_)
    element_type = get_args(type_)[0]
    if element_type is Any:
        return list(list_)
    return [forge(element_type, element) for element in list_]


def forge_mapping(type_: Type[Mapping], mapping: Mapping, forge: Forge) -> Mapping:
    if is_bare(type_):
        return dict(mapping)
    key_arg, value_arg = type_.__args__  # type: ignore
    return {forge(key_arg, k): forge(value_arg, v) for k, v in mapping.items()}



from typing import Any
from typing import List
from typing import Mapping
from typing import Union

import pytest

from svarog.forges import forge_annotated_init
from svarog.forges import forge_list
from svarog.forges import forge_mapping
from svarog.forges import forge_none
from svarog.forges import forge_union
from svarog.types import CannotDispatch
from svarog.types import NoneType


def test_forge_none_should_return_always_none(forge):
    assert forge_none(NoneType, Any, forge) is None


def test_forge_union_should_work_with_optional(forge):
    assert forge_union(Union[str, None], 42, forge) == "42"


def test_forge_union_should_work_with_optional_if_none(forge):
    assert forge_union(Union[str, None], None, forge) is None


def test_forge_list_should_work_with_bare_lists(forge):
    assert forge_list(List, ("foo", 42), forge) == ["foo", 42]


def test_forge_list_should_work_with_non_bare_list(forge):
    assert forge_list(List[int], ("42", "3", "14"), forge) == [42, 3, 14]


@pytest.mark.parametrize(
    "type,data,result",
    [
        (Mapping, {"foo": "bar"}, {"foo": "bar"}),
        (Mapping[str, str], {42: 24}, {"42": "24"}),
        (Mapping[str, Any], {42: 24}, {"42": 24}),
        (Mapping[Any, str], {42: 24}, {42: "24"}),
    ],
)
def test_forge_mapping(forge, type, data, result):
    assert forge_mapping(type, data, forge) == result


def test_forge_annotated_init_should_reject_forging_non_mappings(forge):
    def foo(a):
        assert False

    with pytest.raises(CannotDispatch):
        forge_annotated_init(foo, "non-dict", forge)


def test_forge_annotated_init_should_apply_without_change_if_no_types(forge):
    class A:
        def __init__(self, a):
            self.a = a

    assert forge_annotated_init(A, {"a": 42}, forge).a == 42


def test_forge_annotated_init_should_apply_type_if_not_any(forge):
    class A:
        def __init__(self, a: str):
            self.a = a

    assert forge_annotated_init(A, {"a": 42}, forge).a == "42"



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import pytest

from svarog.checks import has_annotated_init
from svarog.checks import is_bare
from svarog.checks import is_list
from svarog.checks import is_mapping
from svarog.checks import is_union


@pytest.mark.parametrize(
    "type,result", [(Sequence, True), (Dict, False), (Mapping, False)]
)
def test_is_list(type, result):
    assert is_list(type) is result


@pytest.mark.parametrize(
    "type,result",
    [(Optional[str], True), (Union[int, str], True), (Tuple[int, str], False)],
)
def test_is_union(type, result):
    assert is_union(type) is result


class WithAnnotated:
    def __init__(self, a: str):
        ...


@dataclass
class TheDataclass:
    b: int


class WithoutAnnotations:
    def __init__(self, a, b, c):
        ...


@pytest.mark.parametrize(
    "type, result",
    [
        (WithAnnotated, True),
        (TheDataclass, True),
        (WithoutAnnotations, True),
        (str, False),
        (list, False),
    ],
)
def test_has_annotated_init(type, result):
    assert has_annotated_init(type) is result


@pytest.mark.parametrize(
    "type, result",
    [(List, True), (List[Any], False), (Mapping, True), (Mapping[str, Any], False)],
)
def test_is_bare(type, result):
    assert is_bare(type) is result


@pytest.mark.parametrize(
    "type, result", [(List, False), (Mapping, True), (Mapping[str, Any], True)]
)
def test_is_mapping(type, result):
    assert is_mapping(type) is result



from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Any
from typing import ClassVar
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Union
from uuid import UUID

import pytest

from svarog.types import NoneType


def test_can_build_primitive_str(forge):
    assert forge(str, "the-string") == "the-string"


def test_can_build_primitive_int(forge):
    assert forge(int, "3") == 3


def test_can_build_none(forge):
    assert forge(NoneType, None) is None


def test_can_build_dataclass(forge):
    @dataclass
    class A:
        foo: str

    assert forge(A, {"foo": "bar"}) == A(foo="bar")


def test_can_build_complicated_dataclass(forge):
    @dataclass
    class A:
        foo: int
        bar: str
        lorem: Sequence[int]

    @dataclass
    class B:
        a: A
        b: Optional[A]
        c: Sequence[A]

    assert forge(
        B,
        {
            "a": {"foo": "3", "bar": "x", "lorem": [1, 3, 4]},
            "b": None,
            "c": [
                {"foo": 1, "bar": 3, "lorem": ["2", "3"]},
                {"foo": 2, "bar": 3, "lorem": []},
            ],
        },
    ) == B(
        a=A(foo=3, bar="x", lorem=[1, 3, 4]),
        b=None,
        c=[A(1, "3", [2, 3]), A(2, "3", [])],
    )


@dataclass
class A:
    b: "B"
    c: "C"


@dataclass
class B:
    number: int


@dataclass
class C:
    letter: str


def test_can_build_dataclass_from_readme(forge):
    assert forge(A, {"b": {"number": 42}, "c": {"letter": "x"}}) == A(B(42), C("x"))


def test_can_build_dataclass_with_optional_values(forge):
    @dataclass
    class A:
        a: int
        b: Sequence[str] = field(default_factory=lambda: ["x"])

    assert forge(A, {"a": 3}) == A(a=3, b=["x"])


def test_can_build_optional(forge):
    assert forge(Optional[str], None) is None
    assert forge(Optional[str], "str") == "str"


def test_can_build_list(forge):
    assert forge(list, [1, 2, "fooo"]) == [1, 2, "fooo"]


def test_can_build_bare_list(forge):
    assert forge(Sequence, (1, 3, "2")) == [1, 3, "2"]


def test_can_build_list_of_any(forge):
    assert forge(Sequence[Any], (1, 3, "2")) == [1, 3, "2"]


def test_can_build_typed_list(forge):
    assert forge(Sequence[int], (1, 3, "2")) == [1, 3, 2]


def test_can_build_dict(forge):
    assert forge(dict, {"foo": "bar"}) == {"foo": "bar"}


def test_can_build_mapping(forge):
    assert forge(Mapping, {"foo": "bar"}) == {"foo": "bar"}


def test_can_build_typed_mapping(forge):
    assert forge(Mapping[str, int], {"foo": 42, 42: "3"}) == {"foo": 42, "42": 3}


def test_can_build_typed_keys_mapping(forge):
    assert forge(Mapping[str, Any], {"foo": 42, 42: "3"}) == {"foo": 42, "42": "3"}


def test_can_build_typed_value_mapping(forge):
    assert forge(Mapping[Any, int], {"foo": 42, 42: "3"}) == {"foo": 42, 42: 3}


class WithRef:
    def __init__(self, child: Optional["WithRef"]):
        self._child = child

    def __eq__(self, other: "WithRef") -> bool:
        return type(self) == type(other) and self._child == other._child


def test_will_work_with_forward_ref(forge):
    assert forge(WithRef, {"child": {"child": None}}) == WithRef(WithRef(None))


def test_non_optional_unions_are_not_supported_yet(forge):
    with pytest.raises(NotImplementedError):
        assert forge(Union[str, int], "3")


def test_can_apply_simple_types(forge, register_forge):
    @dataclass
    class A:
        uuid: UUID

    assert forge(A, {"uuid": "00000000-0000-0000-0000-000000000000"}) == A(UUID(int=0))


def test_can_register_forge(forge, register_forge):
    class FooType(Enum):
        LOREM = "lorem"
        IPSUM = "ipsum"

    class FooParams:
        types: ClassVar[Mapping[FooType, "FooParams"]] = {}

        def __init_subclass__(cls, type: FooType):
            cls.types[type] = cls

        @classmethod
        def for_type(cls, type):
            return cls.types[type]

    @dataclass
    class LoremFooParams(FooParams, type=FooType.LOREM):
        lorem: str

    @dataclass
    class IpsumFooParams(FooParams, type=FooType.IPSUM):
        ipsum: int

    @dataclass
    class Foo:
        type: FooType
        params: FooParams

        @classmethod
        def forge(cls, _, data, forge):
            foo_type = forge(FooType, data["type"])
            return Foo(
                type=forge(FooType, foo_type),
                params=forge(FooParams.for_type(foo_type), data["params"]),
            )

    register_forge(Foo, Foo.forge)

    assert forge(Foo, {"type": "lorem", "params": {"lorem": "foo-bar"}}) == Foo(
        type=FooType.LOREM, params=LoremFooParams("foo-bar")
    )
    assert forge(Foo, {"type": "ipsum", "params": {"ipsum": 42}}) == Foo(
        type=FooType.IPSUM, params=IpsumFooParams(42)
    )


def test_can_build_class_without_annotations(forge):
    class A:
        def __init__(self, a, b):
            self._a = a
            self._b = b

        def __eq__(self, other):
            return (
                type(self) == type(other)
                and self._a == other._a
                and self._b == other._b
            )

    assert forge(A, {"a": 42, "b": "foo-bar"}) == A(42, "foo-bar")


def test_can_build_if_there_is_default(forge):
    @dataclass
    class A:
        x: int = 0

    assert forge(A, {"x": 42}) == A(42)
